// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// secure is a repository of text security related packages.
package secure // import "golang.org/x/text/secure"
